@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
 /***************************************************************************
  *
- * Copyright (C) 2008 Dan Williams, <dcbw@redhat.com>
+ * Copyright (C) 2008 - 2010 Dan Williams, <dcbw@redhat.com>
  * Copyright (C) 2008 Tambet Ingo, <tambet@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -35,11 +35,13 @@
 #include <glib/gi18n-lib.h>
 #include <gnome-keyring-memory.h>
 #include <nm-setting-connection.h>
+#include <nm-setting-8021x.h>
 
 #include "auth-helpers.h"
 #include "nm-openvpn.h"
 #include "src/nm-openvpn-service.h"
 #include "common-gnome/keyring-helpers.h"
+#include "common/utils.h"
 
 static void
 show_password (GtkToggleButton *togglebutton, GtkEntry *password_entry)
@@ -129,6 +131,47 @@ fill_vpn_passwords (GladeXML *xml,
 	}
 }
 
+static void
+tls_cert_changed_cb (GtkWidget *widget, GtkWidget *next_widget)
+{
+	GtkFileChooser *this, *next;
+	char *fname, *next_fname;
+
+	/* If the just-changed file chooser is a PKCS#12 file, then all of the
+	 * TLS filechoosers have to be PKCS#12.  But if it just changed to something
+	 * other than a PKCS#12 file, then clear out the other file choosers.
+	 *
+	 * Basically, all the choosers have to contain PKCS#12 files, or none of
+	 * them can, because PKCS#12 files contain everything required for the TLS
+	 * connection (CA, client cert, private key).
+	 */
+
+	this = GTK_FILE_CHOOSER (widget);
+	next = GTK_FILE_CHOOSER (next_widget);
+
+	fname = gtk_file_chooser_get_filename (this);
+	if (is_pkcs12 (fname)) {
+		/* Make sure all choosers have this PKCS#12 file */
+		next_fname = gtk_file_chooser_get_filename (next);
+		if (!next_fname || strcmp (fname, next_fname)) {
+			/* Next chooser was different, make it the same as the first */
+			gtk_file_chooser_set_filename (next, fname);
+		}
+		g_free (fname);
+		g_free (next_fname);
+		return;
+	}
+	g_free (fname);
+
+	/* Just-chosen file isn't PKCS#12 or no file was chosen, so clear out other
+	 * file selectors that have PKCS#12 files in them.
+	 */
+	next_fname = gtk_file_chooser_get_filename (next);
+	if (is_pkcs12 (next_fname))
+		gtk_file_chooser_set_filename (next, NULL);
+	g_free (next_fname);
+}
+
 void
 tls_pw_init_auth_widget (GladeXML *xml,
                          GtkSizeGroup *group,
@@ -138,7 +181,7 @@ tls_pw_init_auth_widget (GladeXML *xml,
                          ChangedCallback changed_cb,
                          gpointer user_data)
 {
-	GtkWidget *widget;
+	GtkWidget *widget, *ca, *cert, *key;
 	const char *value;
 	char *tmp;
 	GtkFileFilter *filter;
@@ -149,59 +192,68 @@ tls_pw_init_auth_widget (GladeXML *xml,
 	g_return_if_fail (prefix != NULL);
 
 	tmp = g_strdup_printf ("%s_ca_cert_chooser", prefix);
-	widget = glade_xml_get_widget (xml, tmp);
+	ca = glade_xml_get_widget (xml, tmp);
 	g_free (tmp);
 
-	gtk_size_group_add_widget (group, widget);
-	filter = tls_file_chooser_filter_new ();
-	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
-	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (widget), TRUE);
-	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (widget),
+	gtk_size_group_add_widget (group, ca);
+	if (!strcmp (contype, NM_OPENVPN_CONTYPE_TLS) || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS))
+		filter = tls_file_chooser_filter_new (TRUE);
+	else
+		filter = tls_file_chooser_filter_new (FALSE);
+
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (ca), filter);
+	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (ca), TRUE);
+	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (ca),
 	                                   _("Choose a Certificate Authority certificate..."));
-	g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (changed_cb), user_data);
+	g_signal_connect (G_OBJECT (ca), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 	if (s_vpn) {
 		value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CA);
 		if (value && strlen (value))
-			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value);
+			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (ca), value);
 	}
 
 	if (!strcmp (contype, NM_OPENVPN_CONTYPE_TLS) || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS)) {
 		tmp = g_strdup_printf ("%s_user_cert_chooser", prefix);
-		widget = glade_xml_get_widget (xml, tmp);
+		cert = glade_xml_get_widget (xml, tmp);
 		g_free (tmp);
 
-		gtk_size_group_add_widget (group, widget);
-		filter = tls_file_chooser_filter_new ();
-		gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
-		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (widget), TRUE);
-		gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (widget),
+		gtk_size_group_add_widget (group, cert);
+		filter = tls_file_chooser_filter_new (TRUE);
+		gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (cert), filter);
+		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (cert), TRUE);
+		gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (cert),
 		                                   _("Choose your personal certificate..."));
-		g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (changed_cb), user_data);
+		g_signal_connect (G_OBJECT (cert), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 		if (s_vpn) {
 			value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_CERT);
 			if (value && strlen (value))
-				gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value);
+				gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (cert), value);
 		}
 
 		tmp = g_strdup_printf ("%s_private_key_chooser", prefix);
-		widget = glade_xml_get_widget (xml, tmp);
+		key = glade_xml_get_widget (xml, tmp);
 		g_free (tmp);
 
-		gtk_size_group_add_widget (group, widget);
-		filter = tls_file_chooser_filter_new ();
-		gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
-		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (widget), TRUE);
-		gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (widget),
+		gtk_size_group_add_widget (group, key);
+		filter = tls_file_chooser_filter_new (TRUE);
+		gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (key), filter);
+		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (key), TRUE);
+		gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (key),
 		                                   _("Choose your private key..."));
-		g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (changed_cb), user_data);
+		g_signal_connect (G_OBJECT (key), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 		if (s_vpn) {
 			value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEY);
 			if (value && strlen (value))
-				gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (widget), value);
+				gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (key), value);
 		}
+
+		/* Link choosers to the PKCS#12 changer callback */
+		g_signal_connect (ca, "selection-changed", G_CALLBACK (tls_cert_changed_cb), cert);
+		g_signal_connect (cert, "selection-changed", G_CALLBACK (tls_cert_changed_cb), key);
+		g_signal_connect (key, "selection-changed", G_CALLBACK (tls_cert_changed_cb), ca);
 	}
 
 	if (!strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD) || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS)) {
@@ -317,19 +369,23 @@ validate_file_chooser (GladeXML *xml, const char *name)
 {
 	GtkWidget *widget;
 	char *str;
+	gboolean valid = FALSE;
 
 	widget = glade_xml_get_widget (xml, name);
 	str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
-	if (!str || !strlen (str))
-		return FALSE;
-	return TRUE;
+	if (str && strlen (str))
+		valid = TRUE;
+	g_free (str);
+	return valid;
 }
 
 static gboolean
 validate_tls (GladeXML *xml, const char *prefix, GError **error)
 {
 	char *tmp;
-	gboolean valid;
+	gboolean valid, encrypted = FALSE;
+	GtkWidget *widget;
+	char *str;
 
 	tmp = g_strdup_printf ("%s_ca_cert_chooser", prefix);
 	valid = validate_file_chooser (xml, tmp);
@@ -354,6 +410,7 @@ validate_tls (GladeXML *xml, const char *prefix, GError **error)
 	}
 
 	tmp = g_strdup_printf ("%s_private_key_chooser", prefix);
+	widget = glade_xml_get_widget (xml, tmp);
 	valid = validate_file_chooser (xml, tmp);
 	g_free (tmp);
 	if (!valid) {
@@ -364,6 +421,24 @@ validate_tls (GladeXML *xml, const char *prefix, GError **error)
 		return FALSE;
 	}
 
+	/* Encrypted certificates require a password */
+	str = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
+	encrypted = is_encrypted (str);
+	g_free (str);
+	if (encrypted) {
+		tmp = g_strdup_printf ("%s_private_key_password_entry", prefix);
+		widget = glade_xml_get_widget (xml, tmp);
+		g_free (tmp);
+
+		if (!gtk_entry_get_text_length (GTK_ENTRY (widget))) {
+			g_set_error (error,
+			             OPENVPN_PLUGIN_UI_ERROR,
+			             OPENVPN_PLUGIN_UI_ERROR_INVALID_PROPERTY,
+			             NM_OPENVPN_KEY_CERTPASS);
+			return FALSE;
+		}
+	}
+
 	return TRUE;
 }
 
@@ -461,12 +536,8 @@ update_from_filechooser (GladeXML *xml,
 	g_free (tmp);
 
 	filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (widget));
-	if (!filename)
-		return;
-
-	if (strlen (filename))
+	if (filename && strlen (filename))
 		nm_setting_vpn_add_data_item (s_vpn, key, filename);
-	
 	g_free (filename);
 }
 
@@ -622,6 +693,7 @@ find_tag (const char *tag, const char *buf, gsize len)
 
 static const char *pem_rsa_key_begin = "-----BEGIN RSA PRIVATE KEY-----";
 static const char *pem_dsa_key_begin = "-----BEGIN DSA PRIVATE KEY-----";
+static const char *pem_pkcs8_key_begin = "-----BEGIN ENCRYPTED PRIVATE KEY-----";
 static const char *pem_cert_begin = "-----BEGIN CERTIFICATE-----";
 
 static gboolean
@@ -630,6 +702,7 @@ tls_default_filter (const GtkFileFilterInfo *filter_info, gpointer data)
 	char *contents = NULL, *p, *ext;
 	gsize bytes_read = 0;
 	gboolean show = FALSE;
+	gboolean pkcs_allowed = GPOINTER_TO_UINT (data);
 	struct stat statbuf;
 
 	if (!filter_info->filename)
@@ -642,6 +715,12 @@ tls_default_filter (const GtkFileFilterInfo *filter_info, gpointer data)
 	ext = g_ascii_strdown (p, -1);
 	if (!ext)
 		return FALSE;
+
+	if (pkcs_allowed && !strcmp (ext, ".p12") && is_pkcs12 (filter_info->filename)) {
+		g_free (ext);
+		return TRUE;
+	}
+
 	if (strcmp (ext, ".pem") && strcmp (ext, ".crt") && strcmp (ext, ".key") && strcmp (ext, ".cer")) {
 		g_free (ext);
 		return FALSE;
@@ -676,19 +755,25 @@ tls_default_filter (const GtkFileFilterInfo *filter_info, gpointer data)
 		goto out;
 	}
 
+	if (find_tag (pem_pkcs8_key_begin, (const char *) contents, bytes_read)) {
+		show = TRUE;
+		goto out;
+	}
+
 out:
 	g_free (contents);
 	return show;
 }
 
 GtkFileFilter *
-tls_file_chooser_filter_new (void)
+tls_file_chooser_filter_new (gboolean pkcs_allowed)
 {
 	GtkFileFilter *filter;
 
 	filter = gtk_file_filter_new ();
-	gtk_file_filter_add_custom (filter, GTK_FILE_FILTER_FILENAME, tls_default_filter, NULL, NULL);
-	gtk_file_filter_set_name (filter, _("PEM certificates (*.pem, *.crt, *.key, *.cer)"));
+	gtk_file_filter_add_custom (filter, GTK_FILE_FILTER_FILENAME, tls_default_filter, GUINT_TO_POINTER (pkcs_allowed), NULL);
+	gtk_file_filter_set_name (filter, pkcs_allowed ? _("PEM or PKCS#12 certificates (*.pem, *.crt, *.key, *.cer, *.p12)")
+	                                               : _("PEM certificates (*.pem, *.crt, *.key, *.cer)"));
 	return filter;
 }
 
@@ -755,8 +840,16 @@ sk_file_chooser_filter_new (void)
 static const char *advanced_keys[] = {
 	NM_OPENVPN_KEY_PORT,
 	NM_OPENVPN_KEY_COMP_LZO,
+	NM_OPENVPN_KEY_MSSFIX,
+	NM_OPENVPN_KEY_TUNNEL_MTU,
+	NM_OPENVPN_KEY_FRAGMENT_SIZE,
 	NM_OPENVPN_KEY_TAP_DEV,
 	NM_OPENVPN_KEY_PROTO_TCP,
+	NM_OPENVPN_KEY_PROXY_TYPE,
+	NM_OPENVPN_KEY_PROXY_SERVER,
+	NM_OPENVPN_KEY_PROXY_PORT,
+	NM_OPENVPN_KEY_PROXY_RETRY,
+	NM_OPENVPN_KEY_HTTP_PROXY_USERNAME,
 	NM_OPENVPN_KEY_CIPHER,
 	NM_OPENVPN_KEY_AUTH,
 	NM_OPENVPN_KEY_TA_DIR,
@@ -786,11 +879,21 @@ advanced_dialog_new_hash_from_connection (NMConnection *connection,
 {
 	GHashTable *hash;
 	NMSettingVPN *s_vpn;
+	const char *secret;
 
 	hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
 	s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
 	nm_setting_vpn_foreach_data_item (s_vpn, copy_values, hash);
+
+	/* HTTP Proxy password is special */
+	secret = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+	if (secret) {
+		g_hash_table_insert (hash,
+		                     g_strdup (NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD),
+		                     g_strdup (secret));
+	}
+
 	return hash;
 }
 
@@ -804,6 +907,26 @@ port_toggled_cb (GtkWidget *check, gpointer user_data)
 	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
 }
 
+static void
+tunmtu_toggled_cb (GtkWidget *check, gpointer user_data)
+{
+	GladeXML *xml = (GladeXML *) user_data;
+	GtkWidget *widget;
+
+	widget = glade_xml_get_widget (xml, "tunmtu_spinbutton");
+	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
+}
+
+static void
+fragment_toggled_cb (GtkWidget *check, gpointer user_data)
+{
+	GladeXML *xml = (GladeXML *) user_data;
+	GtkWidget *widget;
+
+	widget = glade_xml_get_widget (xml, "fragment_spinbutton");
+	gtk_widget_set_sensitive (widget, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check)));
+}
+
 static void
 reneg_toggled_cb (GtkWidget *check, gpointer user_data)
 {
@@ -930,6 +1053,11 @@ populate_hmacauth_combo (GtkComboBox *box, const char *hmacauth)
 		NM_OPENVPN_AUTH_NONE,
 		NM_OPENVPN_AUTH_MD5,
 		NM_OPENVPN_AUTH_SHA1,
+		NM_OPENVPN_AUTH_SHA224,
+		NM_OPENVPN_AUTH_SHA256,
+		NM_OPENVPN_AUTH_SHA384,
+		NM_OPENVPN_AUTH_SHA512,
+		NM_OPENVPN_AUTH_RIPEMD160,
 		NULL
 	};
 
@@ -952,6 +1080,16 @@ populate_hmacauth_combo (GtkComboBox *box, const char *hmacauth)
 			name = _("MD-5");
 		else if (!strcmp (*item, NM_OPENVPN_AUTH_SHA1))
 			name = _("SHA-1");
+		else if (!strcmp (*item, NM_OPENVPN_AUTH_SHA224))
+			name = _("SHA-224");
+		else if (!strcmp (*item, NM_OPENVPN_AUTH_SHA256))
+			name = _("SHA-256");
+		else if (!strcmp (*item, NM_OPENVPN_AUTH_SHA384))
+			name = _("SHA-384");
+		else if (!strcmp (*item, NM_OPENVPN_AUTH_SHA512))
+			name = _("SHA-512");
+		else if (!strcmp (*item, NM_OPENVPN_AUTH_RIPEMD160))
+			name = _("RIPEMD-160");
 		else
 			g_assert_not_reached ();
 
@@ -993,6 +1131,55 @@ tls_auth_toggled_cb (GtkWidget *widget, gpointer user_data)
 	gtk_widget_set_sensitive (widget, use_auth);
 }
 
+#define PROXY_TYPE_NONE  0
+#define PROXY_TYPE_HTTP  1
+#define PROXY_TYPE_SOCKS 2
+
+static void
+proxy_type_changed (GtkComboBox *combo, gpointer user_data)
+{
+	GladeXML *xml = GLADE_XML (user_data);
+	gboolean sensitive;
+	GtkWidget *widget;
+	guint32 i = 0;
+	int active;
+	const char *widgets[] = {
+		"proxy_desc_label", "proxy_server_label", "proxy_server_entry",
+		"proxy_port_label", "proxy_port_spinbutton", "proxy_retry_checkbutton",
+		"proxy_username_label", "proxy_password_label", "proxy_username_entry",
+		"proxy_password_entry", NULL
+	};
+	const char *user_pass_widgets[] = {
+		"proxy_username_label", "proxy_password_label", "proxy_username_entry",
+		"proxy_password_entry", NULL
+	};
+
+	active = gtk_combo_box_get_active (combo);
+	sensitive = (active > PROXY_TYPE_NONE);
+
+	while (widgets[i]) {
+		widget = glade_xml_get_widget (xml, widgets[i++]);
+		gtk_widget_set_sensitive (widget, sensitive);
+	}
+
+	/* Additionally user/pass widgets need to be disabled for SOCKS */
+	if (active == PROXY_TYPE_SOCKS) {
+		i = 0;
+		while (user_pass_widgets[i]) {
+			widget = glade_xml_get_widget (xml, user_pass_widgets[i++]);
+			gtk_widget_set_sensitive (widget, FALSE);
+		}
+	}
+
+	/* Proxy options require TCP; but don't reset the TCP checkbutton
+	 * to false when the user disables HTTP proxy; leave it checked.
+	 */
+	widget = glade_xml_get_widget (xml, "tcp_checkbutton");
+	if (sensitive == TRUE)
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+	gtk_widget_set_sensitive (widget, !sensitive);
+}
+
 #define TA_DIR_COL_NAME 0
 #define TA_DIR_COL_NUM 1
 
@@ -1002,8 +1189,11 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	GladeXML *xml;
 	GtkWidget *dialog = NULL;
 	char *glade_file = NULL;
-	GtkWidget *widget;
-	const char *value;
+	GtkWidget *widget, *combo;
+	const char *value, *value2;
+	GtkListStore *store;
+	GtkTreeIter iter;
+	guint32 active = PROXY_TYPE_NONE;
 
 	g_return_val_if_fail (hash != NULL, NULL);
 
@@ -1047,6 +1237,66 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		gtk_widget_set_sensitive (widget, FALSE);
 	}
 
+	/* Proxy support */
+	combo = glade_xml_get_widget (xml, "proxy_type_combo");
+
+	store = gtk_list_store_new (1, G_TYPE_STRING);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("Not required"), -1);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("HTTP"), -1);
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter, 0, _("SOCKS"), -1);
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROXY_SERVER);
+	value2 = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROXY_PORT);
+	if (value && strlen (value) && value2 && strlen (value2)) {
+		long int tmp = 0;
+
+		widget = glade_xml_get_widget (xml, "proxy_server_entry");
+		gtk_entry_set_text (GTK_ENTRY (widget), value);
+
+		errno = 0;
+		tmp = strtol (value2, NULL, 10);
+		if (errno != 0 || tmp < 0 || tmp > 65535)
+			tmp = 0;
+		widget = glade_xml_get_widget (xml, "proxy_port_spinbutton");
+		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
+
+		widget = glade_xml_get_widget (xml, "proxy_retry_checkbutton");
+		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROXY_RETRY);
+		if (value && !strcmp (value, "yes"))
+			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+
+		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME);
+		if (value && strlen (value)) {
+			widget = glade_xml_get_widget (xml, "proxy_username_entry");
+			gtk_entry_set_text (GTK_ENTRY (widget), value);
+		}
+
+		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+		if (value && strlen (value)) {
+			widget = glade_xml_get_widget (xml, "proxy_password_entry");
+			gtk_entry_set_text (GTK_ENTRY (widget), value);
+		}
+	}
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROXY_TYPE);
+	if (value) {
+		if (!strcmp (value, "http"))
+			active = PROXY_TYPE_HTTP;
+		else if (!strcmp (value, "socks"))
+			active = PROXY_TYPE_SOCKS;
+		else
+			active = PROXY_TYPE_NONE;
+	}
+
+	gtk_combo_box_set_model (GTK_COMBO_BOX (combo), GTK_TREE_MODEL (store));
+	g_object_unref (store);
+	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), active);
+	proxy_type_changed (GTK_COMBO_BOX (combo), xml);
+	g_signal_connect (G_OBJECT (combo), "changed", G_CALLBACK (proxy_type_changed), xml);
+
 	widget = glade_xml_get_widget (xml, "port_checkbutton");
 	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (port_toggled_cb), xml);
 
@@ -1072,12 +1322,73 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		gtk_widget_set_sensitive (widget, FALSE);
 	}
 
+	widget = glade_xml_get_widget (xml, "tunmtu_checkbutton");
+	g_assert (widget);
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (tunmtu_toggled_cb), xml);
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TUNNEL_MTU);
+	if (value && strlen (value)) {
+		long int tmp;
+
+		errno = 0;
+		tmp = strtol (value, NULL, 10);
+		if (errno == 0 && tmp > 0 && tmp < 65536) {
+			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+
+			widget = glade_xml_get_widget (xml, "tunmtu_spinbutton");
+			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
+			gtk_widget_set_sensitive (widget, TRUE);
+		}
+	} else {
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
+
+		widget = glade_xml_get_widget (xml, "tunmtu_spinbutton");
+		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 1500.0);
+		gtk_widget_set_sensitive (widget, FALSE);
+	}
+
+	widget = glade_xml_get_widget (xml, "fragment_checkbutton");
+	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (fragment_toggled_cb), xml);
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_FRAGMENT_SIZE);
+	if (value && strlen (value)) {
+		long int tmp;
+
+		errno = 0;
+		tmp = strtol (value, NULL, 10);
+		if (errno == 0 && tmp > 0 && tmp < 65536) {
+			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+
+			widget = glade_xml_get_widget (xml, "fragment_spinbutton");
+			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
+			gtk_widget_set_sensitive (widget, TRUE);
+		}
+	} else {
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
+
+		widget = glade_xml_get_widget (xml, "fragment_spinbutton");
+		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 1300.0);
+		gtk_widget_set_sensitive (widget, FALSE);
+	}
+
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_MSSFIX);
+	if (value && !strcmp (value, "yes")) {
+		widget = glade_xml_get_widget (xml, "mssfix_checkbutton");
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+	}
+
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_COMP_LZO);
 	if (value && !strcmp (value, "yes")) {
 		widget = glade_xml_get_widget (xml, "lzo_checkbutton");
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
 	}
 
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_MSSFIX);
+	if (value && !strcmp (value, "yes")) {
+		widget = glade_xml_get_widget (xml, "mssfix_checkbutton");
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+	}
+
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROTO_TCP);
 	if (value && !strcmp (value, "yes")) {
 		widget = glade_xml_get_widget (xml, "tcp_checkbutton");
@@ -1107,10 +1418,9 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	if (   !strcmp (contype, NM_OPENVPN_CONTYPE_TLS)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD)) {
-		GtkListStore *store;
-		GtkTreeIter iter;
-		int direction = -1, active = -1;
+		int direction = -1;
 
+		active = -1;
 		widget = glade_xml_get_widget (xml, "tls_auth_checkbutton");
 		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TA);
 		if (value && strlen (value))
@@ -1168,7 +1478,8 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 	GtkWidget *widget;
 	GladeXML *xml;
 	const char *contype = NULL;
-	const char *value = NULL;
+	const char *value;
+	int proxy_type = PROXY_TYPE_NONE;
 
 	g_return_val_if_fail (dialog != NULL, NULL);
 	if (error)
@@ -1188,6 +1499,24 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_RENEG_SECONDS), g_strdup_printf ("%d", reneg_seconds));
 	}
 
+	widget = glade_xml_get_widget (xml, "tunmtu_checkbutton");
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+		int tunmtu_size;
+
+		widget = glade_xml_get_widget (xml, "tunmtu_spinbutton");
+		tunmtu_size = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
+		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_TUNNEL_MTU), g_strdup_printf ("%d", tunmtu_size));
+	}
+
+	widget = glade_xml_get_widget (xml, "fragment_checkbutton");
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+		int fragment_size;
+
+		widget = glade_xml_get_widget (xml, "fragment_spinbutton");
+		fragment_size = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
+		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_FRAGMENT_SIZE), g_strdup_printf ("%d", fragment_size));
+	}
+
 	widget = glade_xml_get_widget (xml, "port_checkbutton");
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
 		int port;
@@ -1197,10 +1526,61 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PORT), g_strdup_printf ("%d", port));
 	}
 
+	/* Proxy support */
+	widget = glade_xml_get_widget (xml, "proxy_type_combo");
+	proxy_type = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
+	if (proxy_type != PROXY_TYPE_NONE) {
+		widget = glade_xml_get_widget (xml, "proxy_server_entry");
+		value = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
+		if (value && strlen (value)) {
+			int proxy_port;
+
+			if (proxy_type == PROXY_TYPE_HTTP)
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROXY_TYPE), g_strdup ("http"));
+			else if (proxy_type == PROXY_TYPE_SOCKS)
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROXY_TYPE), g_strdup ("socks"));
+
+			g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROXY_SERVER), g_strdup (value));
+
+			widget = glade_xml_get_widget (xml, "proxy_port_spinbutton");
+			proxy_port = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
+			if (proxy_port > 0) {
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROXY_PORT),
+				                     g_strdup_printf ("%d", proxy_port));
+			}
+
+			widget = glade_xml_get_widget (xml, "proxy_retry_checkbutton");
+			if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROXY_RETRY), g_strdup ("yes"));
+
+			if (proxy_type == PROXY_TYPE_HTTP) {
+				widget = glade_xml_get_widget (xml, "proxy_username_entry");
+				value = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
+				if (value && strlen (value)) {
+					g_hash_table_insert (hash,
+					                     g_strdup (NM_OPENVPN_KEY_HTTP_PROXY_USERNAME),
+					                     g_strdup (value));
+				}
+
+				widget = glade_xml_get_widget (xml, "proxy_password_entry");
+				value = (char *) gtk_entry_get_text (GTK_ENTRY (widget));
+				if (value && strlen (value)) {
+					g_hash_table_insert (hash,
+					                     g_strdup (NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD),
+					                     g_strdup (value));
+				}
+			}
+		}
+	}
+
 	widget = glade_xml_get_widget (xml, "lzo_checkbutton");
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_COMP_LZO), g_strdup ("yes"));
 
+	widget = glade_xml_get_widget (xml, "mssfix_checkbutton");
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
+		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_MSSFIX), g_strdup ("yes"));
+
 	widget = glade_xml_get_widget (xml, "tcp_checkbutton");
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_PROTO_TCP), g_strdup ("yes"));
@@ -1277,3 +1657,24 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 	return hash;
 }
 
+gboolean
+advanced_save_secrets (GHashTable *advanced,
+                       const char *uuid,
+                       const char *name)
+{
+	const char *secret;
+	GnomeKeyringResult result;
+	gboolean ret;
+
+	secret = g_hash_table_lookup (advanced, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+	if (secret && strlen (secret)) {
+		result = keyring_helpers_save_secret (uuid, name, NULL, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD, secret);
+		ret = result == GNOME_KEYRING_RESULT_OK;
+		if (!ret)
+			g_warning ("%s: failed to save HTTP proxy password to keyring.", __func__);
+	} else
+		ret = keyring_helpers_delete_secret (uuid, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+
+	return ret;
+}
+
